// Copyright (c) 2022 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "display/Color.hpp"
#include "state/StateOwner.hpp"

namespace Dawn {
  enum TextureFormat {
    TEXTURE_FORMAT_R = 1,
    TEXTURE_FORMAT_RG = 2,
    TEXTURE_FORMAT_RGB = 3,
    TEXTURE_FORMAT_RGBA = 4
  };

  enum TextureWrapMode {
    TEXTURE_WRAP_MODE_REPEAT = 0,
    TEXTURE_WRAP_MODE_MIRRORED_REPEAT = 1,
    TEXTURE_WRAP_MODE_CLAMP_TO_EDGE = 2,
    TEXTURE_WRAP_MODE_CLAMP_TO_BORDER = 3
  };

  enum TextureFilterMode {
    TEXTURE_FILTER_MODE_NEAREST = 0,
    TEXTURE_FILTER_MODE_LINEAR = 1
  };

  class ITexture : public StateOwner {
    protected:
      bool_t texturePropertiesNeedUpdating = true;

    public:
      StateProperty<enum TextureWrapMode> wrapModeX;
      StateProperty<enum TextureWrapMode> wrapModeY;
      StateProperty<enum TextureFilterMode> filterModeMin;
      StateProperty<enum TextureFilterMode> filterModeMag;
      StateProperty<enum TextureFilterMode> mipmapFilterModeMin;
      StateProperty<enum TextureFilterMode> mipmapFilterModeMag;

      ITexture() :
        wrapModeX(TEXTURE_WRAP_MODE_CLAMP_TO_EDGE),
        wrapModeY(TEXTURE_WRAP_MODE_CLAMP_TO_EDGE),
        filterModeMin(TEXTURE_FILTER_MODE_LINEAR),
        filterModeMag(TEXTURE_FILTER_MODE_LINEAR),
        mipmapFilterModeMin(TEXTURE_FILTER_MODE_NEAREST),
        mipmapFilterModeMag(TEXTURE_FILTER_MODE_NEAREST)
      {
      }

      /**
       * Returns the width of the texture.
       * 
       * @return Width of the texture.
       */
      virtual int32_t getWidth() = 0;

      /**
       * Returns the height of the texture.
       * 
       * @return Height of the texture.
       */
      virtual int32_t getHeight() = 0;

      /**
       * Initializes a texture.
       * 
       * @param width Width of the texture (in pixels).
       * @param height Height of the texture (in pixels).
       * @param format Data format of the texture to use.
       */
      virtual void setSize(
        int32_t width,
        int32_t height,
        enum TextureFormat format
      ) = 0;

      /**
       * Fill a texture with a single color. This is stupidly costly.
       * 
       * @param color Color to fill.
       */
      virtual void fill(struct Color) = 0;
      virtual void fill(uint8_t) = 0;

      /**
       * Returns true only when the texture has been loaded, sized and put on
       * the gpu for rendering.
       * 
       * @return True if ready, otherwise false.
       */
      virtual bool_t isReady() = 0;

      /**
       * Buffer pixel data onto the GPU. Pixel buffering is rather costly so
       * avoid doing this too often.
       * 
       * @param pixels Array of pixels you're trying to buffer.
       * @return The amount of bytes buffered to the texture.
       */
      virtual void buffer(struct ColorU8 pixels[]) = 0;
      virtual void buffer(uint8_t pixels[]) = 0;
  };
}