// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "NewTrueTypeFaceTexture.hpp"

using namespace Dawn;

NewTrueTypeFaceTexture::NewTrueTypeFaceTexture(
  NewTrueTypeFace *trueTypeFace,
  FT_Face &face,
  struct NewTrueTypeFaceTextureStyle style
) {
  assertNotNull(trueTypeFace);
  assertTrue(this->style.fontSize < 256);

  this->trueTypeFace = trueTypeFace;
  this->face = &face;
  this->style = style;

  // Set freetype font size prior to baking.
  if(FT_Set_Pixel_Sizes(face, 0, this->style.fontSize)) {
    assertUnreachable();
  }

  size_t w = 0, h = 0;
  FT_ULong c;

  // First pass, determine the textures' dimensions.
  for(c = NEW_TRUETYPE_CHAR_BEGIN; c < NEW_TRUETYPE_CHAR_END; c++) {
    // Load the character
    if(FT_Load_Char(face, c, FT_LOAD_RENDER | FT_LOAD_MONOCHROME)) {
      assertUnreachable();
    }

    // Update the width and height
    w = mathMax<size_t>(w, face->glyph->bitmap.width);
    h += face->glyph->bitmap.rows;
  }

  assertTrue(w > 0);
  assertTrue(h > 0);

  // Now buffer pixels to the texture
  float_t y = 0;

  // I'd love to just buffer straight to the GPU, but it seems that is a bit 
  // unstable right now.
  uint8_t *buffer = (uint8_t *)memoryFillWithZero(w * h * sizeof(uint8_t));

  size_t offset = 0;
  struct NewTrueTypeCharacter info;
  for(c = NEW_TRUETYPE_CHAR_BEGIN; c < NEW_TRUETYPE_CHAR_END; c++) {
    // Load the character
    if(FT_Load_Char(face, c, FT_LOAD_RENDER)) {
      assertUnreachable();
    }

    // Store the character information
    info.advanceX = face->glyph->advance.x;
    info.advanceY = face->glyph->advance.y;
    info.bitmapSize = glm::vec2(face->glyph->bitmap.width, face->glyph->bitmap.rows);
    info.bitmapPosition = glm::vec2(face->glyph->bitmap_left, -face->glyph->bitmap_top);
    info.textureY = y;
    this->characterData[c] = info;

    // Buffer the pixels, oh dear GOD there has to be a more efficient way.
    for(int32_t i = 0; i < face->glyph->bitmap.rows; i++) {
      memoryCopy(
        (void *)(face->glyph->bitmap.buffer + (i * face->glyph->bitmap.width)),
        (void *)(buffer + offset),
        face->glyph->bitmap.width * sizeof(uint8_t)
      );
      offset += w * sizeof(uint8_t);
      assertTrue(offset <= (w * h * sizeof(uint8_t)));
    }
    y += face->glyph->bitmap.rows;
  }

  this->texture.setSize(w, h, TEXTURE_FORMAT_R);
  this->texture.buffer(buffer);
  memoryFree(buffer);
}

struct NewTrueTypeCharacter NewTrueTypeFaceTexture::getCharacterData(FT_ULong c) {
  return this->characterData[c];
}