// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "NewTrueTypeTool.hpp"

using namespace Dawn;

NewTrueTypeFile::NewTrueTypeFile(std::string path) : file(path) {
  this->path = path;

  style = 0;
  if(path.find("bold") != std::string::npos) {
    style |= NEW_TRUETYPE_VARIANT_BOLD;
  }
  if(path.find("italics") != std::string::npos) {
    style |= NEW_TRUETYPE_VARIANT_ITALICS;
  }

  if(!file.exists()) {
    std::cout << "File " << path << " does not exist!" << std::endl;
    throw "File not found";
  }

  if(!file.open(FILE_MODE_READ)) {
    std::cout << "Failed to open file " << path << " for reading!" << std::endl;
    throw "Unable to open file for reading!";
  }
}

NewTrueTypeFile::~NewTrueTypeFile() {
}


std::vector<std::string> NewTrueTypeTool::getRequiredFlags() {
  return { "output" };
}

std::map<std::string, std::string> NewTrueTypeTool::getOptionalFlags() {
  return {
    { "regular", "" },
    { "italics", "" },
    { "bold", "" },
    { "bold-italics", "" }
  };
}

int32_t NewTrueTypeTool::start() {
  std::vector<NewTrueTypeFile*> files;
  std::vector<std::string> flags = { "regular", "italics", "bold", "bold-italics" };

  auto cleanupFiles = [&]() {
    auto itFile = files.begin();
    while(itFile != files.end()) {
      auto file = *itFile;
      delete file;
      ++itFile;
    }
  };

  // For each flag
  auto itFlag = flags.begin();
  while(itFlag != flags.end()) {
    std::string flag = *itFlag;
    std::string path = this->flags[flag];

    if(path.empty()) {
      ++itFlag;
      continue;
    }

    try {
      auto n = new NewTrueTypeFile(path);
      files.push_back(n);
    } catch(const char *e) {
      cleanupFiles();
      return 1;
    }
    std::cout << "Push back" << std::endl;
    ++itFlag;
  }

  if(files.size() == 0) {
    std::cout << "No valid TTF files provided!" << std::endl;
    cleanupFiles();
    return 1;
  }

  // Create the output file
  File fileOut = File(this->flags["output"]);
  if(!fileOut.mkdirp()) {
    std::cout << "Failed to create output directory!" << std::endl;
    cleanupFiles();
    return 1;
  }
  if(!fileOut.open(FILE_MODE_WRITE)) {
    std::cout << "Failed to open output file for writing!" << std::endl;
    cleanupFiles();
    return 1;
  }
  
  // Prepare to write data
  std::string header = "DE_TTF|3.00|";

  // Write file count
  header += std::to_string(files.size()) + "|";
  
  // For each file
  auto itFile = files.begin();
  while(itFile != files.end()) {
    auto file = *itFile;
    // Style
    header += std::to_string(file->style);
    header += ":";
    // File length
    header += std::to_string(file->file.length);
    header += "|";
    ++itFile;
  }

  if(!fileOut.writeRaw((char *)header.c_str(), header.length())) {
    std::cout << "Failed to write TTF Header to " << fileOut.filename << std::endl;
    cleanupFiles();
    return 1;
  }

  // Now write the data for each file
  itFile = files.begin();
  while(itFile != files.end()) {
    auto file = *itFile;

    // Write the file data
    file->file.setPosition(0);
    if(!fileOut.copyRaw(&file->file, file->file.length)) {
      std::cout << "Failed copy output data of " << file->file.filename << std::endl;
      cleanupFiles();
      return 1;
    }

    // Write vertical bar
    char sep[1];
    sep[0] = '|';
    fileOut.writeRaw(sep, 1);

    ++itFile;
  }

  // Cleanup
  itFile = files.begin();
  while(itFile != files.end()) {
    auto file = *itFile;
    delete file;
    ++itFile;
  }

  // Done
  fileOut.close();
  return 0;
}